import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.Objects;

// Tip podataka koji sadrzi binarno stablo
class BinarnoStablo<T> {

	// Tip podataka koji opisuje jedan cvor stabla
	private static class Cvor<T> {
		
		T element;
		Cvor<T> levo;
		Cvor<T> desno;

		Cvor(T element, Cvor<T> levo, Cvor<T> desno) {
			this.element = element;
			this.levo = levo;
			this.desno = desno;
		}
	}

	// Cvor koji je koren stabla
	// Ako je stablo prazno, koren je null
	protected Cvor<T> koren;

	// Kreiramo prazno stablo
	public BinarnoStablo() {
		koren = null;
	}

	// Kreiramo stablo sa datim elementom u korenu
	// bez levog i desnog podstabla
	public BinarnoStablo(T element) {
		koren = new Cvor<>(element, null, null);
	}

	// Kreiramo novo stablo sa datim elementom u korenu
	// i datim levim i desnim podstablom
	public BinarnoStablo(T element, BinarnoStablo<T> levo, BinarnoStablo<T> desno) {
		koren = new Cvor<T>(element, levo.koren, desno.koren);
	}

	// Zasticeni konstruktor za kreiranje novog stabla
	// sa korenom u datom cvoru ovog stabla
	protected BinarnoStablo(Cvor<T> koren) {
		this.koren = koren;
	}

	public T getElement() {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		return koren.element;
	}

	public void setElement(T element) {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		koren.element = element;
	}

	public BinarnoStablo<T> getLevoPodstablo() {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		if (koren.levo == null) { // Nema levog podstabla
			return null;
		}
		return new BinarnoStablo<T>(koren.levo);
	}

	public void setLevoPodstablo(BinarnoStablo<T> levo) {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		if (levo == null) {
			koren.levo = null;
		} else {
			koren.levo = levo.koren;
		}
	}

	public BinarnoStablo<T> getDesnoPodstablo() {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		if (koren.desno == null) {
			return null;
		}
		return new BinarnoStablo<T>(koren.desno);
	}

	public void setDesnoPodstablo(BinarnoStablo<T> desno) {
		if (koren == null) { // Stablo je prazno
			throw new NoSuchElementException();
		}
		if (desno == null) {
			koren.desno = null;
		} else {
			koren.desno = desno.koren;
		}
	}

	//DatoNaVezbama
	public int brojElemenata() {
		// Spremimo argumente i pozovemo rekurzivni metod
		return brojElemenata(koren);
	}

	// Metod koji zapravo resava dati problem
	//DatoNaVezbama
	protected static <T> int brojElemenata(Cvor<T> cvor) {
		if (cvor == null) {
			return 0;
		}
		int broj = 1;
		broj = broj + brojElemenata(cvor.levo);
		broj = broj + brojElemenata(cvor.desno);
		return broj;
	}

	//DatoNaVezbama
	public int brojNivoa() {
		// Spremimo argumente i pozovemo rekurzivni metod
		return brojNivoa(koren);
	}

	// Metod koji zapravo resava dati problem
	//DatoNaVezbama
	protected static <T> int brojNivoa(Cvor<T> cvor) {
		if (cvor == null) {
			return 0;
		}
		int brojNivoaLevog = brojNivoa(cvor.levo);
		int brojNivoaDesnog = brojNivoa(cvor.desno);
		return Math.max(brojNivoaLevog, brojNivoaDesnog) + 1;
	}

	public boolean jePrazno() {
		return koren == null;
	}

	public boolean postojiElement(T element) {
		return postojiElement(koren, element);
	}
	
	protected static <T> boolean postojiElement(Cvor<T> cvor, T element) {
		if (cvor == null) {
			return false;
		}
		if (Objects.equals(cvor.element, element)) {
			return true;
		}
		if (postojiElement(cvor.levo, element)) {
			return true;
		}
		if (postojiElement(cvor.desno, element)) {
			return true;
		}
		return false;
	}

	public T min(Comparator<T> komparator) {
		return min(koren, komparator);
	}

	protected static <T> T min(Cvor<T> cvor, Comparator<T> komparator) {
		if (cvor == null) {
			return null;
		}
		T min = cvor.element;
		T levoMin = min(cvor.levo, komparator);
		if (komparator.compare(min, levoMin) > 0) {
			min = levoMin;
		}
		T desnoMin = min(cvor.desno, komparator);
		if (komparator.compare(min, desnoMin) > 0) {
			min = desnoMin;
		}
		return min;
	}
	
	public T max(Comparator<T> komparator) {
		return max(koren, komparator);
	}

	protected static <T> T max(Cvor<T> cvor, Comparator<T> komparator) {
		if (cvor == null) {
			return null;
		}
		T max = cvor.element;
		T levoMax = max(cvor.levo, komparator);
		if (komparator.compare(max, levoMax) > 0) {
			max = levoMax;
		}
		T desnoMax = max(cvor.desno, komparator);
		if (komparator.compare(max, desnoMax) > 0) {
			max = desnoMax;
		}
		return max;
	}

	public BinarnoStablo<T> pronadji(T element) {
		Cvor<T> cvor = pronadji(koren, element);
		if (cvor == null) {
			return null;
		}
		return new BinarnoStablo<>(cvor);
	}

	protected static <T> Cvor<T> pronadji(Cvor<T> cvor, T element) {
		if (cvor == null) {
			return null;
		}
		if (Objects.equals(cvor.element, element)) {
			return cvor;
		}
		Cvor<T> nadjenLevo = pronadji(cvor.levo, element);
		if (nadjenLevo != null) {
			return nadjenLevo;
		}
		Cvor<T> nadjenDesno = pronadji(cvor.desno, element);
		if (nadjenDesno != null) {
			return nadjenDesno;
		}
		return null;
	}

	public boolean ubaci(T roditelj, T potomak, boolean levo) {
		return ubaci(koren, roditelj, potomak, levo);
	}

	protected static <T> boolean ubaci(Cvor<T> cvor, T roditelj, T potomak, boolean levo) {
		cvor = pronadji(cvor, roditelj);
		if (cvor == null) {
			return false;
		}
		if (levo) {
			if (cvor.levo != null) {
				return false;
			}
			cvor.levo = new Cvor<T>(potomak, null, null);
		} else {
			if (cvor.desno != null) {
				return false;
			}
			cvor.desno = new Cvor<T>(potomak, null, null);
		}
		return true;
	}

	public BinarnoStablo<T> roditeljOd(T potomak) {
		Cvor<T> cvor = roditeljOd(koren, null, potomak);
		if (cvor == null) {
			return null;
		}
		return new BinarnoStablo<>(cvor);
	}
	
	protected static <T> Cvor<T> roditeljOd(Cvor<T> cvor, Cvor<T> roditelj, T potomak) {
		if (cvor == null) {
			return null;
		}
		if (Objects.equals(cvor.element, potomak)) {
			return roditelj;
		}
		Cvor<T> roditeljLevo = roditeljOd(cvor.levo, cvor, potomak);
		if (roditeljLevo != null) {
			return roditeljLevo;
		}
		Cvor<T> roditeljDesno = roditeljOd(cvor.desno, cvor, potomak);
		if (roditeljDesno != null) {
			return roditeljDesno;
		}
		return null;
	}

	public BinarnoStablo<T> kopija() {
		return new BinarnoStablo<T>(kopija(koren));
	}

	protected static <T> Cvor<T> kopija(Cvor<T> cvor) {
		if (cvor == null) {
			return null;
		}
		return new Cvor<T>(cvor.element, kopija(cvor.levo), kopija(cvor.desno));
	}

	public BinarnoStablo<T> obrnuto() {
		return new BinarnoStablo<T>(obrnuto(koren));
	}

	protected static <T> Cvor<T> obrnuto(Cvor<T> cvor) {
		if (cvor == null) {
			return null;
		}
		return new Cvor<T>(cvor.element, obrnuto(cvor.desno), obrnuto(cvor.levo));
	}

	public void stampajPreorder() {
		stampajPreorder(koren);
	}
	protected static <T> void stampajPreorder(Cvor<T> cvor) {
		if (cvor == null) {
			return;
		}
		Svetovid.out.println(cvor.element);
		stampajPreorder(cvor.levo);
		stampajPreorder(cvor.desno);
	}

	public void stampajInorder() {
		stampajInorder(koren);
	}
	
	protected static <T> void stampajInorder(Cvor<T> cvor) {
		if (cvor == null) {
			return;
		}
		stampajInorder(cvor.levo);
		Svetovid.out.println(cvor.element);
		stampajInorder(cvor.desno);
	}
	
	public void stampajPostorder() {
		stampajPostorder(koren);
	}
	
	protected static <T> void stampajPostorder(Cvor<T> cvor) {
		if (cvor == null) {
			return;
		}
		stampajPostorder(cvor.levo);
		stampajPostorder(cvor.desno);
		Svetovid.out.println(cvor.element);
	}
	
	public void stampajListove() {
		stampajListove(koren);
	}
	
	protected static <T> void stampajListove(Cvor<T> cvor) {
		if (cvor == null) {
			return;
		}
		if ((cvor.levo == null) && (cvor.desno == null)) {
			Svetovid.out.println(cvor.element);
		}
		stampajListove(cvor.levo);
		stampajListove(cvor.desno);
	}
	
	public void stampajSveIspod(T element) {
		stampajSveIspod(koren, element);
	}
	
	protected static <T> void stampajSveIspod(Cvor<T> cvor, T element) {
		Cvor<T> nadjen = pronadji(cvor, element);
		if (nadjen != null) {
			stampajInorder(nadjen.levo);
			stampajInorder(nadjen.desno);
		}
	}
}

	//definisemo neke komparatore za konkretnu klasu koju cemo
	// ubacivati u nase stablo

	class KomparatorPoStarosti implements Comparator<Osoba> {
		public int compare(Osoba o1, Osoba o2) {
			if (o1 == null && o2 == null) {
				return 0;
			}
			if (o1 == null) {
				return 1;
			}
			if (o2 == null) {
				return -1;
			}
			return o2.getGodinaRodjenja() - o1.getGodinaRodjenja();
		}
	}
	
	class KomparatorPoDuziniPrezimena implements Comparator<Osoba> {
		public int compare(Osoba o1, Osoba o2) {
			if (o1 == null && o2 == null) {
				return 0;
			}
			if (o1 == null) {
				return 1;
			}
			if (o2 == null) {
				return -1;
			}
			return o2.getPrezime().length() - o1.getPrezime().length();
		}
	}

// Glavna klasa
public class ResenoStabloProgram {

	// Glavni program
	public static void main(String[] args) {

		// Kreiramo objekat koji koristimo za ucitavanje i ispisivanje stabla
		StabloIOPretty<Osoba> io = new StabloIOPretty<>(Konverter.OSOBA);

		// Ucitamo stablo
		BinarnoStablo<Osoba> stablo = io.readStablo(Svetovid.in("Osobe.txt"));

		// Ispisemo ucitano stablo na ekran
		Svetovid.out.println("Ucitano stablo:");
		io.printStablo(Svetovid.out, stablo);
		Svetovid.out.println();

		// Ilustrujemo pozive metoda
		
		int brojElemenata = stablo.brojElemenata();
		Svetovid.out.println();
		Svetovid.out.println("Broj elemenata je ", brojElemenata);
		
		int brojNivoa = stablo.brojNivoa();
		Svetovid.out.println("Broj nivoa je ", brojNivoa);

		if (stablo.jePrazno()) {
			Svetovid.out.println("Stablo je prazno");
		} else {
			Svetovid.out.println("Stablo nije prazno");
		}
		
		Osoba o1 = new Osoba("Lazar", "Lazarevic", 1976);
		if (stablo.postojiElement(o1)) {
			Svetovid.out.println(o1 + " se nalazi u stablu");
		} else {
			Svetovid.out.println(o1 + " se ne nalazi u stablu");
		}

		Osoba najmladji = stablo.min(new KomparatorPoStarosti());
		Svetovid.out.println("Najmladja osoba je " + najmladji);

		Osoba najduzePrezime = stablo.max(new KomparatorPoDuziniPrezimena());
		Svetovid.out.println("Osoba sa najduzim prezimenom je " + najduzePrezime);

		BinarnoStablo<Osoba> podstablo = stablo.pronadji(o1);
		if (podstablo != null) {
			Svetovid.out.println("Podstablo sa korenom u " + o1 + ":");
			io.printStablo(Svetovid.out, podstablo);
		} else {
			Svetovid.out.println("Podstablo sa korenom u " + o1 + " je prazno");
		}

		Osoba o2 = new Osoba("Krsta", "Krstic", 1988);
		boolean ubacili = stablo.ubaci(o1, o2, true);
		if (ubacili) {
			Svetovid.out.println("Uspesno smo ubacili " + o2 + ":");
		} else {
			Svetovid.out.println("Nismo uspeli da ubacimo novu osobu");
		}

		BinarnoStablo<Osoba> roditelj = stablo.roditeljOd(o1);
		if (roditelj != null) {
			Svetovid.out.println(o1 + " za roditelja ima " + roditelj);
		} else {
			Svetovid.out.println(o1 + " nema roditelja");
		}

		BinarnoStablo<Osoba> kopija = stablo.kopija();
		kopija.setLevoPodstablo(null);
		BinarnoStablo<Osoba> obrnuto = stablo.obrnuto();
		Svetovid.out.println();
		Svetovid.out.println("Kopija stabla (sa odstranjenim levim podstablom):");
		io.printStablo(Svetovid.out, kopija);
		Svetovid.out.println();
		Svetovid.out.println("Obrnuto stablo:");
		io.printStablo(Svetovid.out, obrnuto);
		Svetovid.out.println();
		Svetovid.out.println("Originalno stablo:");
		io.printStablo(Svetovid.out, stablo);

		Svetovid.out.println();
		Svetovid.out.println("Preorder:");
		stablo.stampajPreorder();

		Svetovid.out.println();
		Svetovid.out.println("Inorder:");
		stablo.stampajInorder();

		Svetovid.out.println();
		Svetovid.out.println("Postorder:");
		stablo.stampajPostorder();

		Svetovid.out.println();
		Svetovid.out.println("Listovi:");
		stablo.stampajListove();

		Svetovid.out.println();
		Svetovid.out.println("Svi ispod " + o1 + ":");
		stablo.stampajSveIspod(o1);

	}
}
